# @author Augustin Mortier
# @desc A-Profiles - Clouds detection

from rich.progress import track
from typing import Literal

import aprofiles.profiles
from aprofiles import utils


def detect_clouds(profiles: aprofiles.profiles.ProfilesData, method: Literal["dec", "vg"]="dec", time_avg: float=1., zmin: float=0., thr_noise: float=5., thr_clouds: float=4., min_snr: float=0., verbose: bool=False):
    """Module for *clouds detection*.
    Two methods are available: 
    - "dec" (default): Deep Embedded Clustering (see [AI-Profiles](https://github.com/AugustinMortier/ai-profiles)).
    - "vg": Vertical Gradient.

    Args:
        profiles (aprofiles.profiles.ProfilesData): `ProfilesData` instance.
        method (Literal["dec", "vg"]): cloud detection method used.
        time_avg (float, optional): in minutes, the time during which we aggregates the profiles prior to the clouds detection.
        zmin (float, optional): altitude AGL, in m, above which we look for clouds. We recommend using the same value as used in the extrapolation_low_layers method (only for 'vg' method).
        thr_noise (float, optional): threshold used in the test to determine if a couple (base,peak) is significant: data[peak(z)] data[base(z)] >= thr_noise * noise(z) (only for 'vg' method).
        thr_clouds (float, optional): threshold used to discriminate aerosol from clouds: data[peak(z)] / data[base(z)] >= thr_clouds (only for 'vg' method).
        min_snr (float, optional): Minimum SNR required at the clouds peak value to consider the cloud as valid (only for 'vg' method).
        verbose (bool, optional): verbose mode. Defaults to `False`.

    Returns:
        (aprofiles.profiles.ProfilesData): 
            adds the following (xarray.DataArray) to existing (aprofiles.profiles.ProfilesData):
            - `'clouds' (time, altitude)`: Mask array corresponding to data flagged as a cloud.

    Example 1:
        ```python
        import aprofiles as apro
        # read example file
        path = "examples/data/L2_0-20000-001492_A20210909.nc"
        reader = apro.reader.ReadProfiles(path)
        profiles = reader.read()
        # clouds detection
        profiles.clouds(method="dec")
        # attenuated backscatter image with clouds
        profiles.plot(show_clouds=True, vmin=1e-2, vmax=1e1, log=True)
        ```

        ![Clouds detection](../../assets/images/clouds_dec.png)
    
    Example 2:
        ```python
        import aprofiles as apro
        # read example file
        path = "examples/data/L2_0-20000-001492_A20210909.nc"
        reader = apro.reader.ReadProfiles(path)
        profiles = reader.read()
        # clouds detection
        profiles.clouds(method="vg", zmin=300.)
        # attenuated backscatter image with clouds
        profiles.plot(show_clouds=True, vmin=1e-2, vmax=1e1, log=True)
        ```

        ![Clouds detection](../../assets/images/clouds_vg.png)
    """

    def _vg_clouds(data, zmin, thr_noise, thr_clouds, min_snr):
        # data: 1D range corrected signal (rcs)
        # zmin: altitude AGL, in m, above which we detect clouds
        # thr_noise: threshold used in the test to determine if a couple (base,peak) is significant: data[peak(z)] - data[base(z)] >= thr_noise * noise(z)
        # thr_clouds: threshold used to discriminate aerosol from clouds: data[peak(z)] / data[base(z)] >= thr_clouds
        # min_snr: minimum SNR required at the clouds peak value to consider the cloud as valid. Defaults to 2.

        from scipy import signal
        from scipy.ndimage.filters import uniform_filter1d

        # some useful functions:
        def _get_all_indexes(bases, peaks, tops=[]):
            # get True indexes of bases, peaks and tops, based on masks
            i_bases = utils.get_true_indexes(bases)
            i_peaks = utils.get_true_indexes(peaks)
            i_tops = utils.get_true_indexes(tops)
            return i_bases, i_peaks, i_tops

        def _make_all_masks(data, i_bases, i_peaks, i_tops=[]):
            # return masks for bases, peaks and tops based on input indexes
            bases = utils.make_mask(len(data), i_bases)
            peaks = utils.make_mask(len(data), i_peaks)
            tops = utils.make_mask(len(data), i_tops)
            return bases, peaks, tops

        def _merge_layers(data, i_bases, i_peaks, i_tops):
            # merge layers depending on the altitude of the bases and the tops
            remove_mode = True
            while remove_mode:
                remove_bases, remove_peaks, remove_tops = [], [], []
                for i in range(len(i_bases) - 1):
                    # filters based on the index
                    if i_bases[i + 1] <= i_tops[i]:
                        remove_bases.append(i_bases[i + 1])
                        # remove the weakest peak
                        imin = np.argmin([data[i_peaks[i]], data[i_peaks[i + 1]]])
                        remove_peaks.append(i_peaks[i + imin])
                        # remove lowest top
                        imin = np.argmin([i_tops[i], i_tops[i + 1]])
                        remove_tops.append(i_tops[i + imin])
                        # start again from the bottom
                        break

                i_bases = [i for i in i_bases if i not in remove_bases]
                i_peaks = [i for i in i_peaks if i not in remove_peaks]
                i_tops = [i for i in i_tops if i not in remove_tops]
                if len(remove_bases) == 0:
                    remove_mode = False

            return i_bases, i_peaks, i_tops

        def _find_tops(data, i_bases, i_peaks):
            # function that finds the top of the layers by identifying the first value above thepeak for which the signal is lower than the base
            # if no top is found, the layer is removed
            # retruns lists of indexes of the bases, peaks and tops
            tops = np.asarray([False for x in np.ones(len(data))])
            # conditions: look for bases above i_peaks[i], and data[top[i]] <= data[base[i]]

            i_tops = []
            for i in range(len(i_bases)):
                mask_value = np.asarray(
                    [
                        True if data[j] < data[i_bases[i]] else False
                        for j in range(len(data))
                    ]
                )
                mask_altitude = np.asarray(
                    [True if j > i_peaks[i] else False for j in range(len(data))]
                )
                # the top is the first value that corresponds to the intersection of the two masks
                cross_mask = np.logical_and(mask_value, mask_altitude)
                i_cross_mask = utils.get_true_indexes(cross_mask)
                if len(i_cross_mask) > 0:
                    if tops[i_cross_mask[0]]:
                        bases[i_bases[i]] = False
                        peaks[i_peaks[i]] = False
                    else:
                        tops[i_cross_mask[0]] = True
                        i_tops.append(i_cross_mask[0])
                else:
                    bases[i_bases[i]] = False
                    peaks[i_peaks[i]] = False
            # it is important to keep the tops in the same order, so not to use utils.get_true_indexes() function here
            return utils.get_true_indexes(bases), utils.get_true_indexes(peaks), i_tops

        def _mask_cloud(profile: aprofiles.profiles, bases: list[int], tops: list[int]) -> np.ndarray:
            """Generate a boolean mask for cloud presence in a given profile.

            Args:
                profile (aprofiles.profiles): 1D array representing a single atmospheric profile.
                bases (list[int]): List of base indices where clouds start.
                tops (list[int]): List of top indices where clouds end.

            Returns:
                np.ndarray: Boolean mask of the same length as profile, where cloud regions are True.
            """
            mask = np.zeros_like(profile, dtype=bool)
            
            for base, top in zip(bases, tops):
                mask[base:top] = True
            
            return mask

        #-1. check if any valid data
        if np.isnan(data).all():
            bases, peaks, tops = _make_all_masks(data, [], [], [])
            return {
                "bases": bases,
                "peaks": peaks,
                "tops": tops,
            }

        # 0. rolling average
        avg_data = uniform_filter1d(data, size=10)

        # 1. first derivative
        gradient = np.gradient(avg_data)
        #remove zero values since np.sign(0) = 0
        gradient = [value if value != 0 else 1e-9 for value in gradient]

        # 2. identifies peaks and base by checking the sign changes of the derivative
        sign_changes = np.diff(np.sign(gradient), append=0)
        all_bases = sign_changes == 2
        all_peaks = sign_changes == -2
        # limit to bases above zmin
        imin = profiles._get_index_from_altitude_AGL(zmin)
        all_bases[0:imin] = np.full(imin, False)
        all_peaks[0:imin] = np.full(imin, False)
        # get indexes
        i_bases = utils.get_true_indexes(all_bases)
        i_peaks = utils.get_true_indexes(all_peaks)

        # 3. the signal should start with a base
        if (len(i_bases)>0): #this happens if the whole profile consists of nan
            if i_bases[0] > i_peaks[0] and i_peaks[0] >= 1:
                # set base as the minimum between peak and n gates under
                gates = np.arange(i_peaks[0] - 5, i_peaks[0])
                i_base = gates[np.argmin([data[gates[gates >= 0]]])]
                if i_base >= imin:
                    all_bases[i_base] = True
                else:
                    all_peaks[i_peaks[0]] = False
            # update indexes
            i_bases = utils.get_true_indexes(all_bases)
            i_peaks = utils.get_true_indexes(all_peaks)

        # 4. keeps significant couples (base,peak)
        # a layer can be considered as a proper layer if the difference of signal between the peak and the base is significant (larger than the noise level)
        # noise evaluation (using a high passing frequency filter)
        b, a = signal.butter(1, 0.3, btype="high")
        noise = signal.filtfilt(b, a, data)
        # rolling average of the noise
        avg_abs_noise = uniform_filter1d(abs(noise), size=100)
        # make sure we have as many peaks as bases
        if len(i_peaks) != len(i_bases):
            min_len = min([len(i_peaks), len(i_bases)])
            i_peaks = i_peaks[0:min_len]
            i_bases = i_bases[0:min_len]
        bases, peaks = all_bases, all_peaks
        for i in range(len(i_bases)):
            data_around_peak = avg_data[i_peaks[i]]
            data_around_base = avg_data[i_bases[i]]
            if (
                data_around_peak - data_around_base
                <= thr_noise * avg_abs_noise[i_bases[i]]
            ):
                bases[i_bases[i]] = False
                peaks[i_peaks[i]] = False
        # get indexes
        i_bases = utils.get_true_indexes(bases)
        i_peaks = utils.get_true_indexes(peaks)

        # 5. make sure we finish by a peak: remove last base if necessary
        if len(i_bases) > len(i_peaks):
            bases[i_bases[-1]] = False
            i_bases.pop()

        # 6. find tops of clouds layers
        i_bases, i_peaks, i_tops = _find_tops(avg_data, i_bases, i_peaks)

        # 7. merge layers: for a couple of base and peaks 1,2 if data(b2)>data(p1), then merge layers 1 and 2 by removing p1 and b2
        i_bases, i_peaks, i_tops = _merge_layers(avg_data, i_bases, i_peaks, i_tops)

        # 8. find peaks as maximum between base and top    
        i_peaks = [
            i_bases[i] + np.argmax(data[i_bases[i]: i_tops[i]])
            for i in range(len(i_bases))
        ]
        # reconstruct masks
        bases, peaks, tops = _make_all_masks(data, i_bases, i_peaks, i_tops)

        # 9. distinction between aerosol and clouds
        for i in range(len(i_bases)):
            data_around_peak = avg_data[i_peaks[i]]
            data_around_base = avg_data[i_bases[i]]
            if (
                abs((data_around_peak - data_around_base) / data_around_base)
                <= thr_clouds
            ):
                bases[i_bases[i]] = False
                peaks[i_peaks[i]] = False
                tops[i_tops[i]] = False
        # get indexes
        i_bases, i_peaks, i_tops = _get_all_indexes(bases, peaks, tops)

        """
        #10. set base a closest index
        for i, _ in enumerate(i_peaks):
            mask_value = np.asarray([True if gradient[j]<0 else False for j in range(len(data))])
            mask_altitude = np.asarray([True if j<i_peaks[i] else False for j in range(len(data))])
            #the top is the first value that corresponds to the intersection of the two masks
            cross_mask = np.logical_and(mask_value, mask_altitude)
            i_cross_mask = utils.get_true_indexes(cross_mask)
            i_bases[i] = i_cross_mask[-1]
        """

        # 11. check snr at peak levels
        remove_bases, remove_peaks, remove_tops = [], [], []
        for i in range(len(i_peaks)):
            if utils.snr_at_iz(data, i_peaks[i], step=10) < min_snr:
                remove_bases.append(i_bases[i])
                remove_peaks.append(i_peaks[i])
                remove_tops.append(i_tops[i])
        # remove indexes
        i_bases = [i for i in i_bases if i not in remove_bases]
        i_peaks = [i for i in i_peaks if i not in remove_peaks]
        i_tops = [i for i in i_tops if i not in remove_tops]

        # 11. build cloud mask from indexes
        clouds = _mask_cloud(data, i_bases, i_tops)

        """
        #some plotting
        fig, axs = plt.subplots(1,2,figsize=(10,10))

        ymin, ymax = 000, 15000
        altitude_agl = profiles.data.altitude.data - profiles.data.station_altitude.data

        #signal on the left
        axs[0].plot(data, altitude_agl, 'b', label='rcs')
        axs[0].plot(avg_data, altitude_agl, 'c', label='rcs')
        axs[0].plot(avg_abs_noise,altitude_agl,':b', label='noise level')
        axs[0].plot(avg_abs_noise*thr_noise,altitude_agl,':b', label=f'noise level * {thr_noise}')
        axs[0].plot(data[bases], altitude_agl[bases], '<g', label='bases')
        axs[0].plot(data[peaks], altitude_agl[peaks], '>r', label='peaks')
        axs[0].plot(data[tops], altitude_agl[tops], '^k', label='tops')

        #set axis
        axs[0].set_ylim([ymin, ymax])
        #axs[0].set_xlim([-20000,20000])
        axs[0].legend()

        #derivative on the right
        axs[1].plot(ddata_dz, altitude_agl, 'b', label='first derivative')
        axs[1].plot(ddata_dz[bases], altitude_agl[bases], '<g', label='bases')
        axs[1].plot(ddata_dz[peaks], altitude_agl[peaks], '>r', label='peaks')

        axs[1].set_ylim([ymin, ymax])
        axs[1].legend()
        #set title
        fig.suptitle(t,weight='bold')
        """

        return clouds

    
    def _prepare_data(data, vmin, vmax, target_size):
        # Log transform
        #data = np.log(data)

        # Clip data to the range [vmin, vmax] and then scale it
        data_clipped = np.clip(data, vmin, vmax)

        # Replace NaNs with minimum values
        np.nan_to_num(data_clipped, copy=False, nan=vmin)

        # Invert the normalized data to reflect the 'gray_r' colormap
        data_normalized = (data_clipped - vmin) / (vmax - vmin)
        data_inverted = 1 - data_normalized  # Invert the grayscale values
        
        # Resize the inverted data to the target size
        resized_data = resize(
            data_inverted,
            output_shape=target_size,
            order=0,  # Nearest-neighbor interpolation to avoid smoothing
            anti_aliasing=False
        )
        return resized_data
    
    
    def _ml_clouds(prepared_data, encoder, cluster, target_size, output_shape):
        # Add batch and channel dimensions for further processing
        data_array = np.expand_dims(np.expand_dims(prepared_data, axis=0), axis=-1)

        # Encode the image to get feature representation
        encoded_img = encoder.predict(data_array)[0]  # Remove batch dimension

        # Step 1: Aggregate Encoded Features
        aggregated_encoded_img = np.mean(encoded_img, axis=-1)  # Aggregated to single-channel (16, 32)

        # Optional: Normalize for better visualization
        aggregated_encoded_img = (aggregated_encoded_img - aggregated_encoded_img.min()) / (aggregated_encoded_img.max() - aggregated_encoded_img.min())

        # Step 2: Flatten Encoded Features and Cluster
        encoded_img_flat = encoded_img.reshape(-1, encoded_img.shape[-1])  # Flatten spatial dimensions for clustering
        pixel_labels = cluster.predict(encoded_img_flat)  # Get cluster labels for each pixel

        # Step 3: Map clusters to categories
        category_mapping = {
            1: False,  # molecules
            3: False,  # molecules
            5: False,  # noise
            4: False,  # aerosols
            2: True,  # clouds
            6: True,  # clouds
            0: False,  # other
            7: False   # other
        }
        pixel_labels = np.vectorize(category_mapping.get)(pixel_labels)

        # Reshape the cluster labels back to the spatial dimensions
        pixel_labels_image_shape = pixel_labels.reshape(encoded_img.shape[0], encoded_img.shape[1])

        # Step 4: Upsample the cluster labels to match the original image size
        upsampled_pixel_labels = resize(
            pixel_labels_image_shape,
            (target_size[0], target_size[1]),
            order=0,  # Nearest-neighbor interpolation
            preserve_range=True,
            anti_aliasing=False
        )
        #upsampled_pixel_labels = upsampled_pixel_labels.astype(int)  # Ensure the labels are integers
        
        # resize upsampled_pixel_labels to original size
        resized_upsampled_pixel_labels = resize(
                upsampled_pixel_labels,
                output_shape=output_shape,
                order=0,  # Nearest-neighbor interpolation to avoid smoothing
                anti_aliasing=False
            )
        
        return resized_upsampled_pixel_labels
    
    def _split_matrix(matrix, max_size):
        return [matrix[i:i+max_size] for i in range(0, matrix.shape[0], max_size)]

    def _combine_matrices(matrices):
        return np.vstack(matrices)


    # we work on profiles averaged in time to reduce the noise
    rcs = profiles.time_avg(
        time_avg, var="attenuated_backscatter_0"
    ).data.attenuated_backscatter_0

    # imports and check options
    if method == 'dec':
        import os
        import warnings
        import absl.logging
        from sklearn.exceptions import InconsistentVersionWarning

        # Suppress TensorFlow CUDA messages
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
        os.environ['XLA_FLAGS'] = '--xla_gpu_cuda_data_dir='  # Prevents unnecessary CUDA checks
        os.environ["TF_FORCE_GPU_ALLOW_GROWTH"] = "false"  # Avoids some GPU-related logs

        # Suppress Abseil warnings
        absl.logging.set_verbosity(absl.logging.ERROR)

        # Suppress scikit-learn version warnings
        warnings.simplefilter("ignore", InconsistentVersionWarning)

        # Suppress TensorFlow progress bar
        import tensorflow as tf
        tf.keras.utils.disable_interactive_logging()

        import numpy as np
        from tensorflow.keras.models import load_model
        from skimage.transform import resize
        from pathlib import Path
        import joblib
        
        # ML parameters
        ML = {
            'paths': {
                'encoder': Path(Path(__file__).parent,'ml_models','encoder.keras'),
                'kmeans': Path(Path(__file__).parent,'ml_models','kmeans.pkl')    
            },
            'params': {
                'vmin': -2,
                'vmax': 2,
                'target_size': (256, 512)    
            }
        }
        
        # we split the rcs into subsets, so we artificially increase the time resolution of the cloud detection which is degraded by the DEC technique.
        split_rcs_list = _split_matrix(rcs, 100)
        
        # Load the encoder and kmeans model
        encoder = load_model(ML['paths']['encoder'])
        cluster = joblib.load(ML['paths']['kmeans'])
        
        # prepare data
        split_ml_clouds = []
        for split_rcs in split_rcs_list:
            prepared_data = _prepare_data(split_rcs, ML['params']['vmin'], ML['params']['vmax'], ML['params']['target_size'])
            split_ml_clouds.append(_ml_clouds(prepared_data, encoder, cluster, ML['params']['target_size'], output_shape=np.shape(split_rcs)))
        
        # aggregate split_ml_clouds
        clouds = _combine_matrices(split_ml_clouds)
        
        options = {
            'encoder': ML['paths']['encoder'],
            'kmeans': ML['paths']['kmeans']
        }
        
    elif method == 'vg':
        import numpy as np
        if None in (zmin, thr_noise, thr_clouds, min_snr):
            raise ValueError("For method 'vg', zmin, thr_noise, thr_clouds, and min_snr must be provided.")

        clouds = []
        for i in (track(range(len(profiles.data.time.data)), description="clouds",disable=not verbose)):
            i_clouds = _vg_clouds(
                rcs.data[i, :], zmin, thr_noise, thr_clouds, min_snr
            )

            # store info in 2D array
            clouds.append(i_clouds)
            
        options = {
            'zmin': zmin,
            'thr_noise': thr_noise,
            'thr_clouds': thr_clouds,
            'min_snr': min_snr
        }

    # creates dataarrays
    profiles.data["clouds"] = (("time", "altitude"), clouds)
    profiles.data["clouds"] = profiles.data.clouds.assign_attrs({
        'long_name': 'Clouds mask',
        'units': 'bool',
        'time_avg': time_avg,
        'method': method,
        'options': str(options)
    })
    return profiles


def _main():
    import numpy as np
    import aprofiles as apro

    path = "examples/data/E-PROFILE/L2_0-20000-001492_A20210909.nc"
    
    profiles = apro.reader.ReadProfiles(path).read()

    # basic corrections
    profiles.extrapolate_below(z=150.0, inplace=True)
    #profiles.desaturate_below(z=4000., inplace=True)

    # detection
    profiles.clouds(method="vg", zmin=300, thr_noise=4, thr_clouds=4, verbose=True)
    profiles.plot(show_clouds=True, log=True, vmin=1e-2, vmax=1e1)
    # plot single profile
    #datetime = np.datetime64("2021-09-09T21:20:00")
    #profiles.plot(datetime=datetime, vmin=-1, vmax=10, zmax=12000, show_clouds=True)


if __name__ == "__main__":
    _main()



# @author Augustin Mortier
# @desc A-Profiles - Reference Altitude

import numpy as np
from aprofiles import utils


def get_iref(data, imin, imax, min_snr):
    """Function that returns best index to be used for initializing the Klett backward inversion.

    Args:
        data (1D Array): Attenuated backscatter profile.
        imin (int): Minimum index of the altitude range in which to look for the reference point.
        imax (int): Maximum index of the altitude range in which to look for the reference point.
        min_snr (float): Minimum SNR required to return a valid value.

    Returns:
        (int): index of the reference point.
    """
    # function that returns best index to be used for initializing the Klett backward inversion

    # it is important to copy the data not to modify it outside of the function
    data = data.copy()

    # check if imin below imax
    if imin < imax:
        # limit from imin to imax
        maxdata = np.nanmax(data)
        data[0:imin] = maxdata
        data[imax:] = maxdata

        # running average
        from scipy.ndimage.filters import uniform_filter1d

        avg_data = uniform_filter1d(data, size=3)

        if not np.isnan(np.nanmin(avg_data)):
            # get minimum from the averaged signal
            ilow = np.nanargmin(avg_data)

            # around minimum, get index of closest signal to average signal
            n_around_min = 3
            iclose = np.nanargmin(
                abs(
                    data[ilow - n_around_min: ilow + n_around_min]
                    - avg_data[ilow - n_around_min: ilow + n_around_min]
                )
            )

            if utils.snr_at_iz(data, iclose, step=4) < min_snr:
                return None
            else:
                return ilow + iclose
        else:
            return None
    else:
        return None



# @author Augustin Mortier
# @desc A-Profiles - Aerosol Inversion

import warnings

import json
from pathlib import Path
import numpy as np
from rich.progress import track

import aprofiles as apro

from .ref_altitude import get_iref


def backward_inversion(data, iref, apriori, rayleigh):
    """
    Backward (Klett [Klett1985]_ ) inversion method.

    .. [Klett1985] Klett, J. D. (1985). Lidar inversion with variable backscatter/extinction ratios. Applied optics, 24(11), 1638-1643.

    Args:
        data (array_like): 1D array of a single profile of the attenuated backscatter coefficient.
        iref (float): Index of the reference altitude returned by :func:`aprofiles.retrieval.ref_altitude.get_iref()`.
        apriori (dict): A priori values used to constrain the inversion. Valid keys include:
            - **lr** (float): Lidar Ratio (in sr).
            - **aod** (float): AOD (unitless).
        rayleigh (:class:`aprofiles.rayleigh.RayleighData`): Instance of the `RayleighData` class.

    Raises:
        NotImplementedError: AOD apriori is not implemented yet.

    Returns:
        (array_like): Extinction coefficient (in m⁻¹).
    """

    if "aod" in apriori:
        # search by dichotomy the LR that matches the apriori aod
        raise NotImplementedError("AOD apriori is not implemented yet")
        lr_aer = 50
    else:
        # if we assume the LR, no need to minimize for matching aod
        lr_aer = apriori["lr"]
    lr_mol = 8.0 * np.pi / 3.0

    # vertical resolution
    dz = min(np.diff(rayleigh.altitude))

    int1_a = np.cumsum((lr_aer - lr_mol) * rayleigh.backscatter[:iref] * dz)
    int1_b = [2 * int1_a[-1] - 2 * int1_a[i] for i in range(iref)]
    phi = [np.log(abs(data[i] / data[iref])) + int1_b[i] for i in range(iref)]

    int2_a = 2 * np.nancumsum(lr_aer * np.exp(phi) * dz)
    int2_b = [int2_a[-1] - int2_a[i] for i in range(iref)]

    # initialize total backscatter
    back_aer_iref = 0  # m-1
    beta_tot_iref = rayleigh.backscatter[iref] + back_aer_iref

    # total backscatter
    beta_tot = [np.exp(phi[i]) / ((1 / beta_tot_iref) + int2_b[i]) for i in range(iref)]
    # aerosol backsatter (m-1.sr-1)
    beta_aer = beta_tot - rayleigh.backscatter[:iref]
    # aerosol extinction (m-1)
    sigma_aer = lr_aer * beta_aer
    # returns extinction in m-1 when valid, and np.nan elsewhere
    ext = [sigma_aer[i] if i < len(sigma_aer) else np.nan for i in range(len(data))]
    return ext


def forward_inversion(data, iref, apriori, rayleigh):
    """Forward iterative inversion method [#]_.

    .. [#] Li, D., Wu, Y., Gross, B., & Moshary, F. (2021). Capabilities of an Automatic Lidar Ceilometer to Retrieve Aerosol Characteristics within the Planetary Boundary Layer. Remote Sensing, 13(18), 3626.

    Method principle:

    At z0, the aerosol transmission is assumed as being close to 1. We evaluate the aerosol extinction based on this assumption.
    This evaluation gives a refined aerosol extinction that is used to calculate a second time the aerosol transmission.
    The aerosol extinction retrieval will converge after a certain number iterations.
    After the convergence, the aerosol extinction is retrieved in the next upper layer.

    Args:
        data (array_like): 1D Array of single profile of attenuated backscatter coefficient, in m-1.sr-1.
        iref (float): index of the reference altitude returned by :func:`aprofiles.retrieval.ref_altitude.get_iref()`.
        apriori (dict): A priori value to be used to constrain the inversion. Valid keys: ‘lr’ (Lidar Ratio, in sr) and ‘aod’ (unitless).
        rayleigh (aprofiles.rayleigh.RayleighData). Instance of the :class:`aprofiles.rayleigh.RayleighData` class.

    Raises:
        NotImplementedError: AOD apriori is not implemented yet.

    Returns:
        (array_like): Extinction coefficient (in m⁻¹).
    """

    if "aod" in apriori:
        # search by dichotomy the LR that matches the apriori aod
        raise NotImplementedError("AOD apriori is not implemented yet")
        lr_aer = 50
    elif "lr" in apriori:
        # if we assume the LR, no need to minimize for matching aod
        lr_aer = apriori["lr"]
    lr_mol = 8.0 * np.pi / 3.0

    def _get_aer_at_i(data, i, Tm, Bm, Ta, Ba, Ea, dz, nloop_max=30, diff_ext=0.01):
        # suppress warnings
        warnings.filterwarnings("ignore")

        for _ in range(nloop_max):
            if np.isnan(Ea[i]):
                Ta[i] = 1
            else:
                Ta[i] = np.exp(-np.nancumsum(Ea * dz))[i]
            if (Tm[i] ** 2 * Ta[i] ** 2) != 0:
                Ba[i] = data[i] / (Tm[i] ** 2 * Ta[i] ** 2) - Bm[i]
            else:
                Ba[i] = np.nan
            # test extinction
            if 1 - (lr_aer * Ba[i] / Ea[i]) < diff_ext:
                Ea[i] = lr_aer * Ba[i]
                break
            Ea[i] = lr_aer * Ba[i]

        return Ba[i], Ea[i], Ta[i]

    # vertical resolution
    dz = min(np.diff(rayleigh.altitude))

    Bm = rayleigh.backscatter
    Em = rayleigh.extinction
    Tm = np.exp(-np.cumsum(Em * dz))

    # Initialize aer profiles
    Ta = np.asarray([np.nan for _ in range(len(data))])
    Ba = np.asarray([np.nan for _ in range(len(data))])
    Ea = np.asarray([np.nan for _ in range(len(data))])

    for i in range(iref):
        Ba[i], Ea[i], Ta[i] = _get_aer_at_i(data, i, Tm, Bm, Ta, Ba, Ea, dz)
        if np.isnan(Ba[i]):
            continue
    # returns extinction in m-1
    ext = Ea
    return ext


def inversion(
    profiles,
    time_avg=1,
    zmin=4000.0,
    zmax=6000.0,
    min_snr=0.0,
    under_clouds=False,
    method="forward",
    apriori={"lr": 50.0},
    remove_outliers=False,
    verbose=False,
):
    """Aerosol inversion of the attenuated backscatter profiles using an apriori.

    Args:
        profiles (aprofiles.profiles.ProfilesData): `ProfilesData` instance.
        time_avg (int, optional): in minutes, the time during which we aggregate the profiles before inverting the profiles.
        zmin (float, optional): minimum altitude AGL, in m, for looking for the initialization altitude.
        zmax (float, optional): maximum altitude AGL, in m, for looking for the initialization altitude.
        min_snr (float, optional): Minimum SNR required at the reference altitude to be valid.
        under_clouds (bool, optional): If True, and if the `ProfilesData` has a `cloud_base` variable (returned by the `clouds` method), forces the initialization altitude to be found below the first cloud detected in the profile.
        method (str, optional): must be in [‘backward’, ‘forward’].
        apriori (dict, optional): A priori value to be used to constrain the inversion. Valid keys: ‘lr’ (Lidar Ratio, in sr), ‘aod’ (unitless), 'cfg' (path of config file).
        remove_outliers (bool, optional): Remove profiles considered as outliers based on aod calculation (AOD<0, or AOD>2).
        verbose (bool, optional): verbose mode.

    Raises:
        NotImplementedError: AOD apriori is not implemented yet.

    Returns:
        (aprofiles.profiles.ProfilesData):
            object with additional (xarray.DataArray):

            - `'extinction' (time, altitude)`: 2D array corresponding to the aerosol extinction, in km-1.
            - `'aod' (time)`: 1D array corresponding to the aerosol optical depth associated to the extinction profiles.
            - `'lidar_ratio' (time)`: 1D array corresponding to the lidar ratio associated to the extinction profiles.

    Example:
        Profiles preparation
        ```python
        import aprofiles as apro
        #read example file
        path = "examples/data/L2_0-20000-001492_A20210909.nc"
        reader = apro.reader.ReadProfiles(path)
        profiles = reader.read()
        #extrapolate lowest layers
        profiles.extrapolate_below(z=150, inplace=True)
        ```

        Backward inversion
        ```python
        #aerosol inversion
        profiles.inversion(zmin=4000, zmax=6000, remove_outliers=False, method='backward')
        #plot extinction profiles
        profiles.plot(var='extinction', zmax=6000, vmin=0, vmax=5e-2)
        ```

        ![Extinction profiles retrieved with the backward method](../../assets/images/backward.png)

        Forward inversion
        ```python
        #aerosol inversion
        profiles.inversion(zmin=4000, zmax=6000, remove_outliers=False, method='forward')
        #plot extinction profiles
        profiles.plot(var='extinction', zmax=6000, vmin=0, vmax=5e-2)
        ```

        ![Extinction profiles retrieved with the forward method](../../assets/images/forward.png)
    """

    # we work on profiles averaged in time to reduce the noise
    rcs = profiles.time_avg(
        time_avg, var="attenuated_backscatter_0", inplace=False
    ).data.attenuated_backscatter_0

    """
    #if clouds detected, set to nan profiles where cloud is found below 4000m
    lowest_clouds = profiles._get_lowest_clouds()
    for i in range(len(profiles.data.time.data)):
        if lowest_clouds[i]<=4000:
            rcs.data[i,:] = [np.nan for _ in rcs.data[i,:]]
    """

    # if under_clouds, check if clouds_bases is available
    if under_clouds and "clouds" in list(profiles.data.data_vars):
        lowest_clouds = profiles._get_lowest_clouds()
    elif under_clouds and "clouds" not in list(profiles.data.data_vars):
        warnings.warn(
            "under_clouds parameter sets to True (defaults value) when the clouds detection has not been applied to ProfileData object."
        )
        lowest_clouds = [np.nan for i in np.arange(len(profiles.data.time))]
    else:
        lowest_clouds = [np.nan for i in np.arange(len(profiles.data.time))]

    # aerosol inversion
    ext, lr, aod, z_ref = [], [], [], []
    aod_min, aod_max = 0, 2
    vertical_resolution = profiles._get_resolution("altitude")
    for i in track(
        range(len(profiles.data.time.data)), description="klett ", disable=not verbose
    ):
        altitude_agl = profiles.data.altitude.data
        altitude_asl = (
            profiles.data.altitude.data + profiles.data.station_altitude.data[i]
        )

        # aerosol retrieval requires a molecular profile: use station_altitude in addition
        wavelength = float(profiles.data.l0_wavelength.data)
        rayleigh = apro.rayleigh.RayleighData(
            altitude_asl, T0=298, P0=1013, wavelength=wavelength
        )

        # for this inversion, it is important to use the right units
        if "Mm" in profiles.data.attenuated_backscatter_0.units:
            calibrated_data = (
                rcs.data[i, :] * 1e-6
            )  # conversion from Mm-1.sr-1 to m-1.sr-1
        else:
            calibrated_data = rcs.data[i, :]

        # reference altitude
        imin = profiles._get_index_from_altitude_AGL(zmin)
        imax = profiles._get_index_from_altitude_AGL(
            np.nanmin([zmax, lowest_clouds[i]])
        )
        if method == "backward":
            iref = get_iref(rcs.data[i, :], imin, imax, min_snr)
        elif method == "forward":
            iref = imax
        z_ref.append(altitude_agl[iref])

        if iref is not None:
            # aerosol inversion
            if method == "backward":
                _ext = backward_inversion(calibrated_data, iref, apriori, rayleigh)
            elif method == "forward":
                _ext = forward_inversion(calibrated_data, iref, apriori, rayleigh)
        else:
            _ext = [np.nan for _ in range(len(calibrated_data))]

        # add aod and lr
        if "aod" in apriori:
            # search by dichotomy the LR that matches the apriori aod
            raise NotImplementedError("AOD apriori is not implemented yet")
        else:
            # if we assume the LR, no need to minimize for matching aod
            _lr = apriori["lr"]
            _aod = np.nancumsum(np.asarray(_ext) * vertical_resolution)[-1]
            if remove_outliers and _aod < aod_min or remove_outliers and _aod > aod_max:
                lr.append(np.nan)
                aod.append(np.nan)
                ext.append([np.nan for x in _ext])
            else:
                lr.append(_lr)
                aod.append(_aod)
                ext.append(_ext)

    # creates dataarrays
    profiles.data["extinction"] = (("time", "altitude"), np.asarray(ext) * 1e3)
    profiles.data["extinction"] = profiles.data.extinction.assign_attrs(
        {
            "long_name": f"Extinction Coefficient @ {int(wavelength)} nm",
            "method": f"{method.capitalize()} Klett",
            "units": "km-1",
            "time_avg": time_avg,
            "zmin": zmin,
            "zmax": zmax,
            "apriori_variable": list(apriori.keys())[0],
            "apriori_value": apriori[list(apriori.keys())[0]],
        }
    )

    profiles.data["aod"] = ("time", aod)
    profiles.data["aod"] = profiles.data.aod.assign_attrs(
        {"long_name": f"Aerosol Optical Depth @ {int(wavelength)} nm", "unit": ""}
    )

    profiles.data["lidar_ratio"] = ("time", lr)
    profiles.data["lidar_ratio"] = profiles.data.lidar_ratio.assign_attrs(
        {
            "long_name": f"Lidar Ratio @ {int(wavelength)} nm",
            "units": "sr",
            "use_cfg": str(apriori["use_cfg"]),
        }
    )
    if "cfg" in apriori:
        profiles.data["lidar_ratio"] = profiles.data.lidar_ratio.assign_attrs(
            {"cfg": str(apriori["cfg"])}
        )

    profiles.data["z_ref"] = ("time", z_ref)
    profiles.data["z_ref"] = profiles.data.z_ref.assign_attrs(
        {"long_name": f"Reference altitude ASL", "units": "m"}
    )

    return profiles



# @author Augustin Mortier
# @desc A-Profiles - AERONET reading class

import xarray as xr


class ReadAERONET:
    """AERONET reading class.

    Data:
        [https://data.ceda.ac.uk/badc/eprofile/data/daily_files/]

    Attributes:
        path (str): The path of the file to be read (path or URL).
    """

    def __init__(self, path):
        self.path = path



# @author Augustin Mortier
# @desc A-Profiles - Single profile plot

import matplotlib.pyplot as plt
import numpy as np

import seaborn as sns

sns.set_theme()


def _plot_foc(da, time, zref):
    """Plot foc at the bottom of the image
    Args:
        da ([type]): [description]
        time ([type]): time for which to plot the foc
    """
    # time
    da_time = da.time.data
    i_time = np.argmin(abs(da_time - time))
    # altitude
    if zref.upper() == "AGL":
        altitude = da.altitude.data
    elif zref.upper() == "ASL":
        altitude = da.altitude.data + da.station_altitude.data[i_time]

    foc_markers = [altitude[0] if x else np.nan for x in da.foc.data]
    if not np.isnan(foc_markers[i_time]):
        plt.plot([], [], "^m", ms=10, lw=0, label="fog or condensation")
        plt.plot(0, foc_markers[i_time], "m", marker=10, ms=10, lw=0)


def _plot_clouds(da, time, var, zref):
    """Plot clouds layers
    Args:
        da ([type]): [description]
        time ([type]): time for which to plot the clouds
    """
    # time
    da_time = da.time.data
    i_time = np.argmin(abs(da_time - time))
    # altitude
    if zref.upper() == "AGL":
        altitude = da.altitude.data
    elif zref.upper() == "ASL":
        altitude = da.altitude.data + da.station_altitude.data[i_time]

    c_indexes = da.clouds.data[i_time, :]
    if not np.isnan(da.clouds.data[i_time]).all():
        # plt.plot([], [], "^m", ms=10, lw=0, label=f'Clouds ({da.clouds.method})')
        plt.plot(
            da[var].data[i_time, c_indexes],
            altitude[c_indexes],
            "m",
            marker=10,
            ms=6,
            lw=0,
            label=f"clouds ({da.clouds.method})",
        )


def _plot_pbl(da, time, var, zref):
    """Plot planetary boundary layer
    Args:
        da ([type]): [description]
        time ([type]): time for which to plot the clouds
    """

    # time
    da_time = da.time.data
    i_time = np.argmin(abs(da_time - time))
    # altitude
    if zref.upper() == "AGL":
        altitude = da.altitude.data
    elif zref.upper() == "ASL":
        altitude = da.altitude.data + da.station_altitude.data[i_time]

    # get index of pbl
    i_pbl = np.argmin(abs(da.altitude.data - da.pbl.data[i_time]))

    # plot pbl
    plt.plot(da[var].data[i_time, i_pbl], altitude[i_pbl], "gX", label="PBL")


def plot(
    da,
    datetime,
    var="attenuated_backscatter_0",
    zref="agl",
    zmin=None,
    zmax=None,
    vmin=None,
    vmax=None,
    log=False,
    show_foc=False,
    show_pbl=False,
    show_clouds=False,
    show_fig=True,
    save_fig=None,
):
    """
    Plot single profile of selected variable from (aprofiles.profiles.ProfilesData): object.

    Args:
        da (xarray.DataArray): DataArray.
        datetime (numpy.datetime64): Time for which the profile is plotted.
        var (str, optional): Variable of the DataArray to be plotted.
        zref ({'agl', 'asl'}, optional): Base reference for the altitude axis. Defaults to 'agl'.
        zmin (float, optional): Minimum altitude AGL (m). Defaults to minimum available altitude.
        zmax (float, optional): Maximum altitude AGL (m). Defaults to maximum available altitude.
        vmin (float, optional): Minimum value.
        vmax (float, optional): Maximum value. If None, calculates max from data.
        log (bool, optional): Use logarithmic scale.
        show_foc (bool, optional): Add foc detection.
        show_pbl (bool, optional): Add PBL height.
        show_clouds (bool, optional): Add clouds detection.
        show_fig (bool, optional): Show figure.
        save_fig (str, optional): Path of the saved figure.

    Example:
        ```python
        import aprofiles as apro
        # read example file
        path = "examples/data/L2_0-20000-001492_A20210909.nc"
        reader = apro.reader.ReadProfiles(path)
        profiles = reader.read()
        # some detection
        profiles.clouds(inplace=True).pbl(inplace=True)
        # attenuated backscatter single profile
        datetime = np.datetime64('2021-09-09T10:25:00')
        profiles.plot(datetime=datetime, vmin=-1, vmax=10, zmax=12000, show_clouds=True, show_pbl=True)
        ```

        ![Single profile of attenuated backscatter](../../assets/images/Profile-Oslo-20210909T212005.png){: style="width: 60%;" }
    """

    if datetime is None:
        raise ValueError(
            "datetime needs to be a np.datetime object, e.g. time=np.datetime(2021-09-09T16:00:00)"
        )
    # get index of closest profile
    da_time = da.time.data
    i_time = np.argmin(abs(da_time - datetime))

    # altitude
    if zref.upper() == "AGL":
        altitude = da.altitude.data
    elif zref.upper() == "ASL":
        altitude = da.altitude.data + da.station_altitude.data[i_time]

    fig, axs = plt.subplots(1, 1, figsize=(6, 6))
    plt.plot(da[var].data[i_time], altitude)
    # add zeros
    plt.plot(np.zeros(len(altitude)), altitude, ":k", alpha=0.2)

    if log:
        axs.set_xscale("log")

    # add addition information
    if show_foc:
        _plot_foc(da, da_time[i_time], zref)
    if show_clouds:
        _plot_clouds(da, da_time[i_time], var, zref)
    if show_pbl:
        _plot_pbl(da, da_time[i_time], var, zref)

    # set scales
    plt.ylim([zmin, zmax])
    if vmin is not None or vmax is not None:
        plt.xlim([vmin, vmax])

    # set title and axis labels
    latitude = da.station_latitude.data[i_time]
    longitude = da.station_longitude.data[i_time]
    altitude = da.station_altitude.data[i_time]
    station_id = da.attrs["site_location"]
    # title
    plt.title(
        f"{station_id} ({latitude:.2f};{longitude:.2f};{altitude:.1f}m) - {np.datetime_as_string(da_time[i_time]).split('.')[0]}",
        weight="bold",
        fontsize=12,
    )
    # labels
    if "units" in list(da[var].attrs) and da[var].units is not None:
        xlabel = f"{da[var].long_name} ({da[var].units})"
    else:
        xlabel = f"{da[var].long_name}"
    plt.xlabel(xlabel)
    plt.ylabel(f"Altitude {zref.upper()} (m)")

    # add legend
    if show_foc or show_clouds or show_pbl:
        plt.legend(loc="upper right")

    plt.tight_layout()
    if save_fig:
        plt.savefig(save_fig)
    if show_fig:
        plt.show()



#!/usr/bin/env python
import os
import sys
from pathlib import Path

from glob import glob

import xarray as xr
import pandas as pd
from rich.progress import track

def make_files(path_in: Path, path_out: Path, time_steps: int, progress_bar: bool) -> None:
    
    # list all AP files in path_in
    files = list(Path(path_in).glob("AP*nc"))

    for f in track(files, description=f"Reading AP files", disable=not progress_bar):

        ds = xr.open_dataset(f, decode_times=True, chunks=-1)
        # get unique id and extract yyyymmdd from first time step
        unique_id = f"{ds.attrs['wigos_station_id']}_{ds.attrs['instrument_id']}"
        yyyymmdd = str(ds.time.data[0].astype('M8[D]')).replace('-','')        
        
        # we just work with the 6 latest time steps, that is 30 min worth of data
        start_idx = max(0, ds.time.size - time_steps)
        ds30min = ds.isel(time=slice(start_idx,ds.time.size))
        mmhh = pd.to_datetime(ds30min.time[0].data).strftime('%H%M')
        
        # write output file
        file_name = Path(path_out, f"L2B_{unique_id}{yyyymmdd}{mmhh}.nc")
        ds30min.to_netcdf('out.tmp')
        os.rename('out.tmp',file_name)



# @author Augustin Mortier
# @desc A-Profiles - Profiles writing method

import copy
import os
import warnings

from aprofiles import utils
import numpy as np
import xarray as xr


def write(profiles, base_dir, verbose):
    """
    Writing method for an instance of a (aprofiles.profiles.ProfilesData): class.

    Args:
        profiles (aprofiles.profiles.ProfilesData): Object to be written.
        base_dir (str): Base path of the file should be written.
        verbose (bool): Verbose mode.
    """

    def _classify_scene(ds):
        lowest_clouds = profiles._get_lowest_clouds()
        scene = []
        # got clouds classification here: https://www.metoffice.gov.uk/weather/learn-about/weather/types-of-weather/clouds
        for i, lowest_cloud in enumerate(lowest_clouds):
            if lowest_cloud <= 1981:
                scene.append(3)
            elif lowest_cloud > 1981 and lowest_cloud <= 6096:
                scene.append(2)
            elif lowest_cloud > 6096:
                scene.append(1)
            else:
                scene.append(0)
            # overwrite result based on foc
            if ds.foc.data[i]:
                scene[i] = 4
        return scene

    def _classify_retrieval_scene(ds):
        lowest_clouds = profiles._get_lowest_clouds()
        z_ref = profiles.data.z_ref.data
        scene = []
        for i, _ in enumerate(lowest_clouds):
            if lowest_clouds[i] > z_ref[i]:
                scene.append(1)
            elif lowest_clouds[i] < z_ref[i]:
                scene.append(3)
            else:
                scene.append(0)
            # overwrite result based on foc
            if ds.foc.data[i]:
                scene[i] = 4
        return scene

    # get dataset from profilesdata
    ds = profiles.data

    # get date as string yyyy-mm-dd from first value of the time data
    str_date = str(ds.time.values[0])[:10]
    yyyy = str_date[:4]
    mm = str_date[5:7]
    dd = str_date[8:10]
    filename = f"AP_{ds.wigos_station_id}-{ds.instrument_id}-{str_date}.nc"
    path = os.path.join(base_dir, yyyy, mm, dd, filename)

    if utils.file_exists(path) and verbose:
        warnings.warn(f"{path} already exists and will be overwritten.")
    else:
        from pathlib import Path

        Path(base_dir, yyyy, mm, dd).mkdir(parents=True, exist_ok=True)

    # creates a copy od original dataset -> writes only necessary data
    ds_towrite = copy.deepcopy(ds)

    # for the mass concentration, we just need the mec.
    mec = {}
    for data_var in list(ds_towrite.data_vars):
        if "mass_concentration:" in data_var:
            mec[data_var.split(":")[1]] = ds[data_var].mec
            ds_towrite = ds_towrite.drop(data_var)

    # add mec as new dataarray
    ds_towrite["mec"] = xr.DataArray(
        data=list(mec.values()),
        dims=["aer_type"],
        coords=dict(
            aer_type=list(mec.keys()),
        ),
        attrs=dict(
            long_name="Mass to Extinction Coefficient",
            units="m2.g-1",
            wavelength=ds.l0_wavelength.data,
        ),
    )
    # add attributes to aer_type
    ds_towrite["aer_type"] = ds_towrite["aer_type"].assign_attrs(
        {"long_name": "Aerosol type"}
    )

    # determines the scene classification for each profile
    scene = _classify_scene(ds_towrite)
    # add scene as new dataarray
    ds_towrite["scene"] = ("time", scene)
    ds_towrite["scene"] = ds_towrite["scene"].assign_attrs(
        {
            "long_name": "Scene classification",
            "definition": "0: aer (aerosols only) / 1: high_cloud (base cloud above 6096 m) / 2: mid_cloud (base cloud between 1981 m and 6096 m) / 3: low-cloud (base cloud below 1981 m) - 4: foc (fog or condensation)",
        }
    )

    # add scene for extinction profile: cloud_above, cloud_below
    retrieval_scene = _classify_retrieval_scene(ds_towrite)
    # add scene as new dataarray
    ds_towrite["retrieval_scene"] = ("time", retrieval_scene)
    ds_towrite["retrieval_scene"] = ds_towrite["retrieval_scene"].assign_attrs(
        {
            "long_name": "Retrieval scene classification",
            "definition": "0: aer (aerosols only) / 1: cloud_above (cloud above reference altitude) / 3: cloud_below (cloud below reference point) / 4: foc (fog or condensation)",
        }
    )

    # drop other variables
    drop_variables = [
        "start_time",
        "vertical_visibility",
        "cbh_uncertainties",
        "uncertainties_att_backscatter_0",
    ]
    for drop_var in drop_variables:
        ds_towrite = ds_towrite.drop(drop_var)

    # some variables have no dimension. Set it as attribute and drop the variable.
    nodim_variables = ["l0_wavelength"]
    for nodim_var in nodim_variables:
        ds_towrite.attrs[nodim_var] = ds_towrite[nodim_var].data
        ds_towrite = ds_towrite.drop(nodim_var)

    # for convenience, add initial coordinates of station
    coordinates_vars = ["station_altitude", "station_latitude", "station_longitude"]
    for coordinate_var in coordinates_vars:
        ds_towrite.attrs[f"{coordinate_var}_t0"] = ds_towrite[coordinate_var].data[0]

    # add altitude direction
    ds_towrite["altitude"] = ds_towrite["altitude"].assign_attrs({"positive": "up"})

    # encoding with compression
    encoding = {}
    for varname, var in ds_towrite.variables.items():
        if varname == "time":
            continue
        if var.dtype == np.int64:
            encoding[varname] = {
                "dtype": np.int32,
                "zlib": True,
                "chunksizes": var.shape,
            }
        if varname in ["extinction", "clouds"]:
            encoding[varname] = {"zlib": True, "chunksizes": var.shape}

    # convert also the quality_flag's variable flag_values attribute also to NC_INT instead of NC_INT64
    ds_towrite["quality_flag"] = ds_towrite.quality_flag.assign_attrs(
        {"flag_values": np.array([0, 1, 2], dtype=np.int32)}
    )

    # writes to netcdf
    # ds_towrite.to_netcdf(path, mode="w", encoding=encoding)
    ds_towrite.to_netcdf(path, mode="w")



# @author Augustin Mortier
# @desc A-Profiles - Mass to Extinction Coefficient
import json
from pathlib import Path

import matplotlib.pyplot as plt
import miepython
import numpy as np

from aprofiles import size_distribution


class MECData:
    """
    Class for computing the *Mass to Extinction Coefficient* for a given aerosol type.
    This class calls the [`get_mec()`](#aprofiles.mec.MECData.get_mec) method.

    Attributes:
       `aer_type` ({'dust', 'volcanic_ash', 'biomass_burning', 'urban'}): aerosol type.
       `wavelength` (int or float): wavelength, in mm.
       `method` ({'mortier_2013', 'literature'}): method to retrieve or compute `MEC`.
       `aer_properties` (dict): dictionary describing the optical and micro-physical properties of the prescribed aerosol (read from *aer_properties.json*)

    Example:
        ```python
        #some imports
        import aprofiles as apro
        mec_data = MECData('volcanic_ash', 532.)
        mec_data.__dict__.keys()
        dict_keys(['aer_type', 'wavelength', 'aer_properties', 'nsd', 'vsd', 'radius', 'qext', 'conv_factor', 'mec'])
        print(f'{mec_data.conv_factor:.2e} m {mec_data.mec):.2e} m2.g-1')
        6.21e-07 m 0.62 m2.g-1
        ```
    """

    def __init__(self, aer_type: str, wavelength: float, method: str = "mortier_2013"):

        # check parameters type
        if not isinstance(aer_type, str):
            raise TypeError(
                "`aer_type` is expected to be a string. Got {} instead.".format(
                    type(aer_type)
                )
            )
        if not isinstance(wavelength, (int, float)):
            raise TypeError(
                "`wavelength` is expected to be an int or a float. Got {} instead".format(
                    type(wavelength)
                )
            )
        available_methods = ["mortier_2013", "literature"]
        if method not in available_methods:
            raise ValueError(
                "Invalid `method`. AAvailable methods are {}".format(available_methods)
            )

        self.aer_type = aer_type
        self.wavelength = wavelength
        self.method = method

        if self.method == "mortier_2013":
            # read aer_properties.json files
            f = open(Path(Path(__file__).parent,'config','aer_properties.json'))
            aer_properties = json.load(f)
            f.close()
            # check if the aer_type exist in the json file
            if aer_type not in aer_properties.keys():
                raise ValueError(
                    "{} not found in aer_properties.json. `aer_type` must be one of the follwowing: {}".format(
                        aer_type, list(aer_properties.keys())
                    )
                )
            else:
                self.aer_properties = aer_properties[self.aer_type]
                self.get_mec()
        elif self.method == "literature":
            self.mec = 3.33  # CHECK V-PROFILES VALUES. CHECK CODE IN LAPTOP?
            self.conv_factor = -99

    def get_mec(self):
        """
        Calculates the Extinction to Mass Coefficient for a given type of particles, assuming a prescribed size distribution shape (with unknown amplitude), density, and using [Mie theory](https://miepython.readthedocs.io) to calculate the extinction efficiency.
        
        Returns:
            (MECData): with additional attributes:
            
                - `nsd` (1D Array): Number Size Distribution
                - `vsd` (1D Array): Volume Size Distribution
                - `radius` (1D Array): Radius in µm
                - `x` (1D Array): Size parameter (unitless)
                - `conv_factor` (float): Conversion factor in m
                - `mec` (float): Extinction to Mass Coefficient in m².g⁻¹

       !!! note
            For a population of particles, the extinction coefficient $\sigma_{ext}$ (m⁻¹) can be written as follows:

            $$
            \sigma_{ext} = \int_{r_{min}}^{r_{max}} N(r) Q_{ext}(m, r, \lambda) \pi r^2 dr
            $$

            where $Q_{ext}$ is the `extinction efficiency` and $N(r)$ is the `Number Size Distribution` (NSD).

            $Q_{ext}$ varies with the refractive index, $m$, the wavelength, $\lambda$, and can be calculated for spherical particles using Mie theory.

            The total aerosol mass concentration $M_0$ (µg.m⁻³) can be expressed as:

            $$
            M_0 = \int_{r_{min}}^{r_{max}} M(r) dr
            $$

            where $M(r)$ is the mass size distribution (MSD).

            This can be rewritten in terms of NSD and MSD as:

            $$
            M_0 = \int_{r_{min}}^{r_{max}} \\frac{4\pi r^3}{3} \\rho N(r) dr
            $$

            where $\\rho$ is the particle density (kg.m⁻³).

            By normalizing the NSD with respect to the fine mode ($N(r) = N_0 N_1(r)$), we arrive at:

            $$
            M_0 = \sigma_{ext} \\rho \\frac{4}{3} \\frac{\int_{r_{min}}^{r_{max}} N_1(r) r^3 dr}{\int_{r_{min}}^{r_{max}} N_1(r) Q_{ext}(m, r, \lambda) r^2 dr}
            $$

            We define the `conversion factor` (in m) as:

            $$
            c_v = \\frac{4}{3} \\frac{\int_{r_{min}}^{r_{max}} N_1(r) r^3 dr}{\int_{r_{min}}^{r_{max}} N_1(r) Q_{ext}(m, r, \lambda) r^2 dr}
            $$

            Thus, the equation simplifies to:

            $$
            M_0 = \sigma_{ext} \\rho c_v
            $$

            Finally, the `Extinction to Mass Coefficient` (MEC, also called `mass extinction cross section`, usually in m².g⁻¹) is defined as:

            $$
            MEC = \\frac{\sigma_{ext}}{M_0} = \\frac{1}{\\rho c_v}
            $$

            with $\\rho$ expressed in g.m⁻³.

            
            <table>
                <thead>
                    <tr>
                        <th>Aerosol Type</th>
                        <th colspan=2>Conversion Factor (µm)</th>
                        <th colspan=2>MEC (m².g⁻¹) </th>
                    </tr>
                    <tr>
                        <th></th>
                        <th>532 nm</th>
                        <th>1064 nm</th>
                        <th>532 nm</th>
                        <th>1064 nm</th>
                    </tr>
                </thead>
                <tbody>
                    <tr>
                        <td>Urban</td>
                        <td>0.31</td>
                        <td>1.92</td>
                        <td>1.86</td>
                        <td>0.31</td>
                    </tr>
                    <tr>
                        <td>Desert dust</td>
                        <td>0.68</td>
                        <td>1.04</td>
                        <td>0.58</td>
                        <td>0.38</td>
                    </tr>
                    <tr>
                        <td>Biomass burning</td>
                        <td>0.26</td>
                        <td>1.28</td>
                        <td>3.30</td>
                        <td>0.68</td>
                    </tr>
                    <tr>
                        <td>Volcanic ash</td>
                        <td>0.62</td>
                        <td>0.56</td>
                        <td>0.62</td>
                        <td>0.68</td>
                    </tr>
                </tbody>
                <caption>Conversion Factors and MEC calculated for the main aerosol types</caption>
            </table>

        """


        def _compute_conv_factor(nsd, qext, radius):
            """Compute Conversion Factor for a given Size Distribution and Efficiency

            Args:
                nsd (1D Array): Number Size Distribution (µm-3.µm-1)
                qext (1D Array): Extinction Efficiency (unitless)
                radius (1D Array): Radius, in µm

            Returns:
                (float): Conversion factor (m)

            """
            # radius resolution
            dr = min(np.diff(radius))

            # integrals
            numerator = [nsd[i] * (radius[i] ** 3) for i in range(len(radius))]
            denominator = [
                nsd[i] * qext[i] * (radius[i] ** 2) for i in range(len(radius))
            ]
            int1 = np.nancumsum(np.asarray(numerator) * dr)[-1]
            int2 = np.nancumsum(np.asarray(denominator) * dr)[-1]

            conv_factor = (4 / 3) * (int1 / int2)

            # conversion form µm to m
            conv_factor = conv_factor * 1e-6
            return conv_factor

        # generate a size distribution for given aer_type
        sd = size_distribution.SizeDistributionData(self.aer_type)

        # calculate efficiency extinction qext
        # size parameter
        # as the radius is in µm and the wavelength is in nm, one must convert the wavelength to µm
        x = [2 * np.pi * r / (self.wavelength * 1e-3) for r in sd.radius]
        # refractive index
        m = complex(
            self.aer_properties["ref_index"]["real"],
            -abs(self.aer_properties["ref_index"]["imag"]),
        )
        # mie calculation
        qext, _qsca, _qback, _g = miepython.mie(m, x)

        # output
        self.nsd = sd.nsd
        self.vsd = sd.vsd
        self.radius = sd.radius
        self.x = x
        self.qext = qext
        self.conv_factor = _compute_conv_factor(sd.nsd, qext, sd.radius)
        self.mec = 1 / (
            self.conv_factor * self.aer_properties["density"] * 1e6
        )  # convert density from g.cm-3 to g.m-3
        return self

    def plot(self, show_fig=True):
        """
        Plot main information of an instance of the [`MECData`](#aprofiles.mec.MECData) class.

        Example:
            ```python
            #import aprofiles
            import aprofiles as apro
            #compute mec for biomas burning particles at 532 nm
            mec = apro.mec.MECData('volcanic_ash', 532.);
            #plot information
            mec.plot()
            ```

            ![Volcanic Ash particles properties used for MEC calculation](../../assets/images/volcanic_ash-mec.png)
        """
        fig, ax = plt.subplots(1, 1, figsize=(6, 6))

        # plot Volume Size Distribution in 1st axis
        ax.plot(self.radius, self.vsd, label="VSD")
        ax.set_ylabel("V(r) (µm2.µm-3)")

        # plot Number Size Distribution in 2nd axis
        if "nsd" in self.__dict__:
            # add secondary yaxis
            ax2 = ax.twinx()
            ax2.plot(self.radius, self.qext, label="Qext", color="gray")
            ax2.set_ylabel("Qext (unitless)")
            # ax2.set_ylim([0,10])

        # add additional information
        plt.text(
            0.975,
            0.85,
            f"$at\ \lambda={self.wavelength:.0f}\ nm$",
            horizontalalignment="right",
            verticalalignment="center",
            transform=ax.transAxes,
        )
        plt.text(
            0.975,
            0.80,
            f"$c_v: {self.conv_factor * 1e6:.2f}\ \mu m$",
            horizontalalignment="right",
            verticalalignment="center",
            transform=ax.transAxes,
        )
        plt.text(
            0.975,
            0.75,
            f"$MEC: {self.mec:.2f}\ m^2/g$",
            horizontalalignment="right",
            verticalalignment="center",
            transform=ax.transAxes,
        )

        ax.set_xlabel("Radius (µm)")
        ax.set_xscale("log")
        fig.legend(
            loc="upper right", bbox_to_anchor=(1, 1), bbox_transform=ax.transAxes
        )
        plt.title(
            f"{self.aer_type.capitalize().replace('_', ' ')} particles properties for MEC calculation",
            weight="bold",
        )
        plt.tight_layout()
        if show_fig:
            plt.show()


def _main():
    import aprofiles as aprofiles

    mec_data = MECData("urban", 1064.0)
    print(f"{mec_data.conv_factor:.2e} m {mec_data.mec:.2f} m2.g-1")
    mec_data.plot()


if __name__ == "__main__":
    _main()



# @author Augustin Mortier
# @desc A-Profiles - Fog or Condensation detection

from aprofiles import utils
import numpy as np
from typing import Literal

def _detect_fog_from_cloud(profiles, zmin_cloud):
    # returns a bool list with True where fog/condensation cases
    # if the base of the first cloud (given by the constructor) is below
    first_cloud_base_height = profiles._get_lowest_clouds()
    foc = [True if x <= zmin_cloud else False for x in first_cloud_base_height]
    return foc

def _detect_fog_from_snr(profiles, z_snr, var, min_snr):
    # returns a bool list with True where fog/condensation cases

    # calculates snr at z_snr
    iz_snr = profiles._get_index_from_altitude_AGL(z_snr)
    # calculates snr at each timestamp
    snr = [
        utils.snr_at_iz(profiles.data[var].data[i, :], iz_snr, step=4)
        for i in range(len(profiles.data.time.data))
    ]
    # condition
    foc = [True if x <= min_snr else False for x in snr]
    return foc

def detect_foc(profiles, method: Literal["cloud_base", "snr"]="cloud_base", var="attenuated_backscatter_0", z_snr=2000., min_snr=2., zmin_cloud=200.):
    """Detects fog or condensation. Two methods are available:
    - cloud_base (default): uses the minimum cloud base height.
    - snr (fallback if no clouds are available): identifies areas under a certain altitude for which snr values lower than the prescribed threshold.

    Args:
        profiles (aprofiles.profiles.ProfilesData): `ProfilesData` instance.
        method ({'cloud_base', 'snr'}, optional): Detection method.
        var (str, optional): Used for 'snr' method. Variable to calculate SNR from.
        z_snr (float, optional): Used for 'snr' method. Altitude AGL (in m) at which we calculate the SNR.
        min_snr (float, optional): Used for 'snr' method. Minimum SNR under which the profile is considered as containing fog or condensation.
        zmin_cloud (float, optional): Used for 'cloud_base' method. Altitude AGL (in m) below which a cloud base height is considered a fog or condensation situation.

    Returns:
        (aprofiles.profiles.ProfilesData): 
            adds the following (xarray.DataArray) to existing (aprofiles.profiles.ProfilesData):
            
            - `'foc' (time)`: mask array corresponding to the presence of foc.

    Example:
        ```python
        import aprofiles as apro
        # read example file
        path = "examples/data/L2_0-20000-001492_A20210909.nc"
        reader = apro.reader.ReadProfiles(path)
        profiles = reader.read()
        # foc detection
        profiles.foc()
        # attenuated backscatter image with pbl up to 6km of altitude
        profiles.plot(show_foc=True, zmax=6000., vmin=1e-2, vmax=1e1, log=True)
        ```

        ![Fog or condensation (foc) detection](../../assets/images/foc.png)
    """

    if method == "cloud_base" and "clouds" in profiles.data:
        foc = _detect_fog_from_cloud(profiles, zmin_cloud)
    else:
        foc = _detect_fog_from_snr(profiles, z_snr, var, min_snr)

    # creates dataarray
    profiles.data["foc"] = ("time", foc)
    profiles.data["foc"] = profiles.data.foc.assign_attrs({'long_name': 'Fog or condensation mask'})

    return profiles


def _main():
    import aprofiles as apro

    path = "examples/data/E-PROFILE/L2_0-20000-001492_A20210909.nc"
    profiles = apro.reader.ReadProfiles(path).read()

    # basic corrections
    profiles.extrapolate_below(z=150.0, inplace=True)
    # detection
    profiles.foc().plot(show_foc=True, log=True, vmin=1e-2, vmax=1e1)


if __name__ == "__main__":
    _main()



# @author Augustin Mortier
# @desc A-Profiles - E-PROFILE reading class

import numpy as np
import xarray as xr


class ReadEPROFILE:
    """
    E-PROFILE reading class.

    Data:
        [https://data.ceda.ac.uk/badc/eprofile/data/daily_files/]

    Attributes:
        path (str): The path of the file to be read (path or URL).
    """

    def __init__(self, path):
        self.path = path

    def _check_path(self):
        """
        Check the path

        Raises:
            OSError: [description]
            NotImplementedError: [description]
        """
        # check if file is NetCDF
        filename = self.path.split("/")[-1]
        if not filename.endswith(".nc"):
            raise OSError(
                f"NetCDF file required (.nc). File format not supported {filename}"
            )
        # check if web address
        if self.path.startswith("https://"):
            raise NotImplementedError(
                "The reading of CEDA Archive is not implemented yet"
            )

    def _read_l2file(self):
        """
        Read single file using `xr.open_dataset()`.

        Returns:
            (xarray.Dataset):
        """
        ds = xr.open_dataset(self.path, decode_times=True)
        # remove time duplicate values if exists
        if len(ds.time.values) != len(np.unique(ds.time.values)):
            ds = ds.drop_duplicates(dim="time")
        # in CEDA archive, dimensions come as (altitude, time). Transpose all variables which have altitude as dimension.
        if ds.latitude.dims[0] == "altitude":
            ds = ds.transpose(..., "altitude")

        # make sure that the station_altitude variable follows the time dimension
        position_keys = ["station_latitude", "station_longitude", "station_altitude"]
        for position_key in position_keys:
            if type(ds[position_key].data.tolist()) == float:
                if (
                    position_key == "station_altitude"
                    and ds.altitude.long_name == "Altitude above sea level"
                ):
                    # Store existing attributes
                    altitude_attrs = ds.altitude.attrs
                    ds = ds.assign_coords(
                        altitude=np.round(ds.altitude - ds[position_key].data, 3)
                    )
                    altitude_attrs["long_name"] = "Altitude above ground level"
                    ds["altitude"] = ds.altitude.assign_attrs(altitude_attrs)

                altitude_array = np.ones(np.shape(ds.time.data)) * ds[position_key].data
                ds[position_key] = xr.DataArray(
                    data=altitude_array, dims=["time"], coords={"time": ds.time}
                )

        # replace wavelength with actual value in attenuated backscatter longname
        ds.attenuated_backscatter_0.attrs["long_name"] = (
            ds.attenuated_backscatter_0.long_name.replace(
                "at wavelength 0", f"@ {int(ds.l0_wavelength.data)} nm"
            )
        )
        ds.attenuated_backscatter_0.attrs["units"] = ds.attenuated_backscatter_0.attrs[
            "units"
        ].replace("1E-6*1/(m*sr)", ("Mm-1.sr-1"))
        return ds

    def read(self):
        """
        Method which reads E-PROFILE data.

        Returns:
            (xarray.Dataset):
        """

        self._check_path()

        # check if the filename starts with L2
        filename = self.path.split("/")[-1]
        if filename.startswith("L2"):
            return self._read_l2file()
        else:
            raise NotImplementedError("Only L2 files are supported at the moment")



# @author Augustin Mortier
# @desc A-Profiles Rayleigh Profile

import matplotlib.pyplot as plt
import numpy as np


class RayleighData:
    """
    Class for computing *rayleigh profile* in a standard atmosphere.
    This class calls the :func:`get_optics_in_std_atmo()` method, which produces profiles of `backscatter` and `extinction` coefficients.

    Attributes:
        altitude (array-like): array of altitude ASL to be used to compute the rayleigh profile, in m.
        wavelength (float): Wavelength of the Rayleigh profile to be computed, in nm.
        T0 (float): Temperature at ground level, in K.
        P0 (float): Pressure at ground level, in hPa.

    Example:
        ```python
        # some imports
        import aprofiles as apro
        import numpy as np
        # creates altitude array
        altitude = np.arange(15,15000,15)
        wavelength = 1064.
        # produce rayleigh profile
        rayleigh = apro.rayleigh.RayleighData(altitude, wavelength, T0=298, P0=1013);
        # checkout the instance attributes
        rayleigh.__dict__.keys()
        dict_keys(['altitude', 'T0', 'P0', 'wavelength', 'cross_section', 'backscatter', 'extinction'])
        ```
    """

    def __init__(self, altitude: list, wavelength: float, T0=298, P0=1013):
        self.altitude = altitude
        self.T0 = T0
        self.P0 = P0
        self.wavelength = wavelength

        # calls functions
        self.get_optics_in_std_atmo()

    def get_optics_in_std_atmo(self):
        """
        Function that returns *backscatter* and *extinction* profiles [#]_ for an instance of a (RayleighData): class.

        .. [#] Bucholtz, A. (1995). Rayleigh-scattering calculations for the terrestrial atmosphere. Applied optics, 34(15), 2765-2773.

        Returns:
            (aprofiles.rayleigh.RayleighData): object with additional attributes:

                - `extinction` (numpy.ndarray): extinction coefficient (m-1)
                - `backscatter` (numpy.ndarray): backscatter coefficient (m-1.sr-1)
                - `cross_section` (numpy.ndarray): cross section (cm-2)
        """

        def _refi_air(wavelength):
            """
            Function that calculates the refractive index of the air at a given wavelength in a standard atmosphere [#]_.

            .. [#] Peck, E. R., & Reeder, K. (1972). Dispersion of air. JOSA, 62(8), 958-962.

            Args:
                wavelength (float): wavelength, in µm.

            Returns:
                refractive index of the air.
            """
            # wav (float): wavelength in µm
            # returns refractive index of the air in standard atmosphere (Peck and Reeder)
            var = (
                8060.51
                + 2480990 / (132.274 - wavelength ** -2)
                + 17455.7 / (39.32957 - wavelength ** -2)
            )
            return var * 1e-8 + 1

        # standard values at sea level
        T0 = 298  # K
        P0 = 1013 * 1e2  # Pa
        p_He = 8

        # standard gradients & parameters
        atmo = {
            "tropo": {"zmin": 0, "zmax": 13, "dTdz": -6.5},
            "strato": {"zmin": 13, "zmax": 55, "dTdz": 1.4},
            "meso": {"zmin": 55, "zmax": 100, "dTdz": -2.4},
        }

        # convert altitude to km
        z = self.altitude / 1000

        # vertical resolution
        dz = min(z[1:] - z[0:-1])

        # temperature profile
        Tz = [T0]
        for layer in atmo.keys():
            ibottom = int(np.floor(atmo[layer]["zmin"] / dz))
            itop = int(np.floor(atmo[layer]["zmax"] / dz))
            for i in range(ibottom, itop):
                Tz.append(Tz[ibottom] + (i - ibottom) * atmo[layer]["dTdz"] * dz)

        # vertical coordinates for temperature
        z_Tz = np.arange(0, 100, dz)

        # pressure profile
        Pz = P0 * np.exp(-z_Tz / p_He)

        # molecules density and cross section
        Ns = (6.02214e23 / 22.4141) / 1000  # molecules.cm-3
        # density (cm-3)
        N_m = [Ns * (T0 / P0) * (Pz[i] / Tz[i]) for i in range(len(Pz))]

        # cross section, in cm2 (adapted from Bodhaine et al., 1999)
        king_factor = 1.05  # tomasi et al., 2005
        num = 24 * (np.pi ** 3) * ((_refi_air(self.wavelength * 1e-3) ** 2 - 1) ** 2)
        denum = (
            ((self.wavelength * 1e-7) ** 4)
            * (Ns ** 2)
            * ((_refi_air(self.wavelength * 1e-3) ** 2 + 2) ** 2)
        )
        section_m = (num / denum) * king_factor

        # extinction profile
        amol = N_m * np.array(section_m)
        # backscatter profile
        lr_mol = 8 * np.pi / 3
        bmol = amol / lr_mol

        # colocate vertically to input altitude
        imin = np.argmin(np.abs(np.array(z_Tz) - z[0]))
        imax = imin + len(z)

        # output
        self.cross_section = section_m  # in cm2
        self.backscatter = bmol[imin:imax] * 1e2  # from cm-1 to m-1
        self.extinction = amol[imin:imax] * 1e2  # from cm-1 to m-1
        self.tau = np.cumsum(self.extinction*(dz*1000))[-1]

        return self

    def plot(self, show_fig=True):
        """
        Plot extinction profile of an instance of the (RayleighData): class.

        Example:
            ```python
            # some imports
            import aprofiles as apro
            import numpy as np
            # creates altitude array
            altitude = np.arange(15,15000,15)
            wavelength = 1064.
            # produce rayleigh profile
            rayleigh = apro.rayleigh.RayleighData(altitude, wavelength, T0=298, P0=1013);
            # plot profile
            rayleigh.plot()
            ```

            .. figure:: ../../docs/assets/images/rayleigh.png
                :scale: 80 %
                :alt: rayleigh profile

                Rayleigh extinction profile for a standard atmosphere.
        """

        fig, ax = plt.subplots(1, 1, figsize=(6, 6))
        plt.plot(self.extinction, self.altitude)
        plt.text(
            0.97,
            0.94,
            f"$\sigma_m: {self.cross_section:.2e} cm2$",
            horizontalalignment="right",
            verticalalignment="center",
            transform=ax.transAxes,
        )
        plt.text(
            0.97,
            0.88,
            f"$\u03C4_m: {self.tau:.2e}$",
            horizontalalignment="right",
            verticalalignment="center",
            transform=ax.transAxes,
        )
        
        plt.title(
            f"Rayleigh Profile in a Standard Atmosphere ({self.P0}hPa, {self.T0}K)",
            weight="bold",
        )
        plt.xlabel(f"Extinction coefficient @ {self.wavelength} nm (m-1)")
        plt.ylabel("Altitude ASL (m)")
        plt.tight_layout()
        if show_fig:
            plt.show()


def _main():
    import aprofiles as apro

    altitude = np.arange(15, 15000, 15)
    wavelength = 1064.0
    rayleigh = RayleighData(altitude, wavelength)
    # plot
    rayleigh.plot()


if __name__ == "__main__":
    _main()



# @author Augustin Mortier
# @desc A-Profiles - Code for creating calendar json file to be used in V-Profiles

import json
from pathlib import Path

import xarray as xr


def make_calendar(path, yyyy: str, mm: str, calname: str) -> None:
    # one calendar, per month
    with open(Path(path, yyyy, mm, calname), 'w') as json_file:
        json.dump({}, json_file)

def add_to_calendar(fn, path, yyyy: str, mm: str, dd: str, calname: str) -> None:
    # calendar collects the number of inversions with no low-level clouds (<6km) at each station
    # for each station, write number of each scene class (aer, cloud<6km, cloud>6km, )
    
    # read data
    vars_to_read = ['retrieval_scene']
    
    # in some cases, the file might be corrupted. Just skip it then.
    try:
        ds = xr.open_dataset(fn, engine='netcdf4', chunks=-1)[vars_to_read].load()
    except OSError:
        print(f'File {fn} is corrupted. Skipping...')
        return

    # counts scenes
    scene_classes = [4, 3, 1, 0]
    scene_counts = {}
    for scene_class in scene_classes:
        scene_counts[scene_class] = len(ds.retrieval_scene.data[ds.retrieval_scene.data == scene_class])
    scene_counts['total'] = len(ds.retrieval_scene.data)

    # open current calendar
    with open(Path(path, yyyy, mm, calname), 'r') as json_file:
        data = json.load(json_file)
    json_file.close()        

    # add new data to calendar data
    station_id = f'{ds.attrs["wigos_station_id"]}-{ds.attrs["instrument_id"]}'
    if station_id not in data:
        data[station_id] = {}
    data[station_id][dd] = scene_counts

    # write new calendar
    with open(Path(path, yyyy, mm, calname), 'w') as json_file:
        json.dump(data, json_file)



# @author Augustin Mortier
# @desc A-Profiles - Utils

import numpy as np
import os

def get_true_indexes(mask):
    """
    Returns indexes of a list of boolean values where the values are True.

    Args:
        - mask (list): list of Bool.
    """
    return [i for i, x in enumerate(mask) if x]

def make_mask(length, indexes_where_True):
    """
    Returns a list of Bool of a given length with True values at given indexes.

    Args:
        - length (int): length of the list.
        - indexes_where_True (list): indexes where the values are True.
    """
    # length: int: length of the mask
    # indexes_where_true: list
    mask = np.asarray([False for x in np.ones(length)])
    mask[indexes_where_True] = len(indexes_where_True) * [True]
    return mask

def file_exists(path):
    """
    Returns True is file exists.

    Args:
        - path (str): path of the file to be checked.
    """
    return os.path.exists(path)

def snr_at_iz(array, iz, step):
    """
    Returns the SNR from an array calculated at iz with step points above and below.

    Args:
        - array (list): data array.
        - iz (int): index where to calculate SNR.
        - step (int): number of steps to be used for SNR calculation.
    """
    # calculates the snr from array at iz around step points
    gates = np.arange(iz - step, iz + step)
    indexes = [i for i in gates if i > 0 and i < len(array)]
    mean = np.nanmean(array[indexes])
    std = np.nanstd(array[indexes], ddof=0)
    if std != 0:
        return mean / std
    else:
        return 0

def gaussian(x, mu, sig, norm):
    """1D Gaussian function

    Args:
        x (1D-Array): array of values for which to return the Gaussian function.
        mu (float): Mean
        sig (float): Standard deviation
        norm (float): Normalization factor

    Returns:
        1D-Array: Gaussian distribution
    """
    return (norm / (np.sqrt(2 * np.pi) * sig)) * np.exp(
        -((x - mu) ** 2 / (2 * sig ** 2))
    )


# @author Augustin Mortier
# @desc A-Profiles - Read config file

import json
from pathlib import Path


def read():
    # read alc_parameters files
    with open(Path(Path(__file__).parent, '..', 'config', 'cfg.json')) as json_file:
        return json.load(json_file)



# @author Augustin Mortier
# @desc A-Profiles - Reader class

from aprofiles.io import read_aeronet, read_eprofile
from aprofiles.profiles import ProfilesData


class ReadProfiles:
    """
    Base class for reading profiles data.

    Attributes:
        path (str): path of the file to be read.
    """

    def __init__(self, path):
        self.path = path

    def read(self):
        """
        Method which calls the relevant reading class based on the file name.

        Returns:
            (aprofiles.profiles.ProfilesData):

        Example:
            ```python
            # import library
            import aprofiles as apro
            path = "examples/data/E-PROFILE/L2_0-20000-001492_A20210909.nc"
            # initialize reading class with file path
            reader = apro.reader.ReadProfiles(path)
            # calls the read method
            profiles = reader.read()
            # print ProfilesData object
            print(profiles)
            <aprofiles.profiles.ProfilesData object at 0x7f011055fad0>
            # print the content of this ProfilesData object
            profiles.__dict__
            {'_data': <xarray.Dataset>
            Dimensions:                          (altitude: 511, layer: 3, time: 273)
            Coordinates:
            * time                             (time) datetime64[ns] 2021-09-09T00:00:04 ... 2021-09-09T23:55:06
            * altitude                         (altitude) float64 111.0 ... 1.541e+04
            Dimensions without coordinates: layer
            Data variables:
                start_time                       (time) datetime64[ns] ...
                latitude                         (time, altitude) float64 ...
                longitude                        (time, altitude) float64 ...
                attenuated_backscatter_0         (time, altitude) float64 ...
                uncertainties_att_backscatter_0  (time, altitude) float64 ...
                l0_wavelength                    float64 1.064e+03
                station_longitude                float64 ...
                station_latitude                 float64 ...
                station_altitude                 float64 ...
                quality_flag                     (time, altitude) int64 ...
                vertical_visibility              (time) float64 ...
                cloud_base_height                (time, layer) float64 ...
                cbh_uncertainties                (time, layer) float64 ...
                cloud_amount                     (time) float64 ...
                calibration_constant_0           (time) float64 ...
            Attributes:
                instrument_id:                A
                hermes_history:               hermes-raw2l1 2.0.1, hermes-eprofile-alc-sc...
                instrument_type:              CHM15k
                site_location:                OSLO,NORWAY
                title:                        OSLO nimbus MET NORWAY
                principal_investigator:       Remote Sensing Group
                optical_module_id:            TUB110019
                comment:                      
                Conventions:                  CF-1.7, UKMO-1.0.2
                wigos_station_id:             0-20000-0-01492
                source:                       Ground Based Remote Sensing
                references:                   E-PROFILE Data Format Description Document
                wmo_id:                       01492
                overlap_is_corrected:         true
                instrument_firmware_version:  1.04
                overlap_function:             true
                institution:                  MET NORWAY Remote Sensing Group
                instrument_serial_number:     TUB110019
                history:                      Thu Sep  9 00:30:28 2021: ncpdq -O -a time,...
                NCO:                          netCDF Operators version 4.9.1 (Homepage = ...}
            ```
        """

        # get the right reading class
        data = read_eprofile.ReadEPROFILE(self.path).read()

        # instantitate ProfilesData class with data
        profiles = ProfilesData(data)

        return profiles


class ReadAeronet:
    """
    Base class for reading Aeronet data.

    Attributes:
        path (str): path of the file to be read.
    """

    def __init__(self, path):
        self.path = path
        raise NotImplementedError("ReadAeronet class is not implemented yet")


def _main():
    path = "examples/data/E-PROFILE/L2_0-20000-001492_A20210909.nc"
    profiles = ReadProfiles(path).read()
    print(profiles)
    print(profiles.data.time)


if __name__ == "__main__":
    _main()



# @author Augustin Mortier
# @desc A-Profiles - Image plot

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import numpy as np
import pandas as pd

import seaborn as sns

sns.set_theme()


def _plot_foc(da, zref):
    """Plot foc at the bottom of the image
    Args:
        da ([type]): [description]
    """
    # time
    time = da.time.data
    # altitude
    if zref.upper() == "AGL":
        altitude = da.altitude.data
    else:
        raise ValueError("Unsupported altitude reference. Use 'AGL'.")

    foc_markers = [altitude[0] if x else np.nan for x in da.foc.data]

    # plot line from peak to base
    for i, _ in enumerate(foc_markers):
        if not np.isnan(foc_markers[i]):
            plt.plot(
                [time[i], time[i]],
                [altitude[0], altitude[-1]],
                "-",
                color="black",
                lw=2.0,
                alpha=0.4,
            )
    # plot markers
    plt.plot([], [], "^m", ms=10, lw=0, label="fog or condensation")
    plt.plot(time, foc_markers, "m", marker=10, ms=10, lw=0)


def _plot_clouds(da, zref):
    """Plot clouds.
    Args:
        da ([type]): [description]
    """
    # time
    time = da.time.data
    clouds = da.clouds
    # altitude
    if zref.upper() == "AGL":
        altitude = da.altitude.data
    else:
        raise ValueError("Unsupported altitude reference. Use 'AGL'.")

    # 2D array
    C = np.transpose(clouds.data)
    C_plot = np.where(C, 1, np.nan)

    plt.pcolormesh(
        time,
        altitude,
        C_plot,
        shading="nearest",
        cmap="Greys_r",
        vmin=0,
        vmax=1,
        alpha=0.9,
    )

    # Manually create a legend entry
    plt.plot([], [], lw=0, marker="s", ms=10, color="white", alpha=0.9, label="clouds")


def _plot_pbl(da, zref):
    """Plot PBL as markers
    Args:
        da ([type]): [description]
    """
    # time
    time = da.time.data
    # altitude
    if zref.upper() == "AGL":
        pbl = da.pbl.data
    else:
        raise ValueError("Unsupported altitude reference. Use 'AGL'.")
    plt.plot(time, pbl, ".g", ms=5, lw=0, label="PBL")


def plot(
    da,
    var="attenuated_backscatter_0",
    zref="agl",
    zmin=None,
    zmax=None,
    vmin=None,
    vmax=None,
    log=False,
    show_foc=False,
    show_pbl=False,
    show_clouds=False,
    cmap="coolwarm",
    show_fig=True,
    save_fig=None,
):
    """Plot image of selected variable from :class:`aprofiles.profiles.ProfilesData` object.

    Args:
        da (xarray.DataArray): DataArray.
        var (str, optional): Variable of the DataArray to be plotted.
        zref ({'agl'}, optional): Base reference for altitude axis.
        zmin (float, optional): Minimum altitude AGL (m).
        zmax (float, optional): Maximum altitude AGL (m).
        vmin (float, optional): Minimum value.
        vmax (float, optional): Maximum value. If `None`, calculates max from data.
        log (bool, optional): Use logarithmic scale.
        show_foc (bool, optional): Add foc detection.
        show_pbl (bool, optional): Add PBL height.
        show_clouds (bool, optional): Add clouds detection.
        cmap (str, optional): Matplotlib colormap.
        show_fig (bool, optional): Show figure.
        save_fig (str, optional): Path of the saved figure.

    Example:
        ``` python
        import aprofiles as apro
        # read example file
        path = "examples/data/L2_0-20000-001492_A20210909.nc"
        reader = apro.reader.ReadProfiles(path)
        profiles = reader.read()
        # attenuated backscatter image
        profiles.plot(vmin=1e-2, vmax=1e1, log=True)
        ```
        ![Image of attenuated backscatter profiles](../../assets/images/attenuated_backscatter.png)
    """

    # calculates max value from data
    if vmax is None and type(da[var].data[0][0]) != np.bool:
        perc = np.percentile(da[var].data, 70)
        pow10 = np.ceil(np.log10(perc))
        vmax = 10 ** (pow10)

    # time
    time = da.time.data
    # altitude
    if zref.upper() == "AGL":
        altitude = da.altitude.data
    else:
        raise ValueError("Unsupported altitude reference. Use 'AGL'.")

    fig, axs = plt.subplots(1, 1, figsize=(12, 4))

    # 2D array
    C = da[var].data.T

    if log:
        import matplotlib.colors as colors

        pcm = plt.pcolormesh(
            time,
            altitude,
            C,
            norm=colors.LogNorm(vmin=np.max([0, vmin]), vmax=vmax),
            cmap=cmap,
            shading="nearest",
        )
    else:
        pcm = plt.pcolormesh(
            time, altitude, C, vmin=vmin, vmax=vmax, cmap=cmap, shading="nearest"
        )
    # store colorbar
    cbar = fig.colorbar(pcm, ax=axs)

    # add addition information
    if show_foc:
        _plot_foc(da, zref)
    if show_clouds:
        _plot_clouds(da, zref)
    if show_pbl:
        _plot_pbl(da, zref)

    # limit to altitude range
    plt.ylim([zmin, zmax])

    # set title and axis labels
    yyyy = pd.to_datetime(da.time.values[0]).year
    mm = pd.to_datetime(da.time.values[0]).month
    dd = pd.to_datetime(da.time.values[0]).day
    latitude = da.station_latitude.data
    longitude = da.station_longitude.data
    altitude = da.station_altitude.data
    station_id = da.attrs["site_location"]
    # title
    str_latitude = (
        f"{latitude[0]:.2f}"
        if len(np.unique(latitude)) == 1
        else f"{np.min(latitude):.2f}-{np.max(latitude):.2f}"
    )
    str_longitude = (
        f"{longitude[0]:.2f}"
        if len(np.unique(longitude)) == 1
        else f"{np.min(longitude):.2f}-{np.max(longitude):.2f}"
    )
    str_altitude = (
        f"{altitude[0]:.2f}"
        if len(np.unique(altitude)) == 1
        else f"{np.min(altitude):.2f}-{np.max(altitude):.2f}"
    )

    plt.title(
        f"{station_id} ({str_latitude};{str_longitude};{str_altitude}m) - {yyyy}/{mm:02}/{dd:02}",
        weight="bold",
    )
    # labels
    plt.xlabel("Time")
    plt.ylabel(f"Altitude {zref.upper()} (m)")

    # add legend
    if show_foc or show_clouds or show_pbl:
        plt.legend(loc="upper right")

    # label
    if "units" in list(da[var].attrs) and da[var].units is not None:
        label = f"{da[var].long_name} ({da[var].units})"
    else:
        label = f"{da[var].long_name}"
    cbar.set_label(label)

    plt.tight_layout()
    if save_fig:
        plt.savefig(save_fig)
    if show_fig:
        plt.show()



#!/usr/bin/env python3
# @author Augustin Mortier
# @desc A-Profiles - CLI for running aprofiles standard workflow

import concurrent.futures
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import List

import typer
from pandas import date_range
from rich.progress import Progress, track

import aprofiles.cli.utils as utils

app = typer.Typer(no_args_is_help=True)

class InstrumentType(str, Enum):
    chm15k = "CHM15k"
    minimpl = "Mini-MPL"
    cl61 = "CL61"

@app.command()
def run(
    _dates: List[datetime] = typer.Option(
        [], "--date", formats=["%Y-%m-%d"], help="📅 Processing date."
    ),
    _from: datetime = typer.Option(
        None, "--from", formats=["%Y-%m-%d"], help="📅 Initial date."
    ),
    _to: datetime = typer.Option(
        datetime.today(),
        "--to",
        formats=["%Y-%m-%d"],
        help="📅 Ending date.",
        show_default="Today's date",
    ),
    today: bool = typer.Option(False, help="🕑 Process today's data."),
    yesterday: bool = typer.Option(False, help="🕙 Process yesterday's data."),
    instruments_types: List[InstrumentType] = typer.Option(
        ['CHM15k', 'Mini-MPL'], "--instruments-type", help="📗 List of specific instruments to be processed."
    ),
    multiprocessing: bool = typer.Option(False, help="🚀 Use multiprocessing mode."),
    workers: int = typer.Option(
        2, "--workers", min=1, envvar="NSLOTS", help="👷 Number of workers (NSLOTS, if multiprocessing mode is enabled)."
    ),
    path_in: Path = typer.Option(
        "data/e-profile", exists=True, readable=True, help="📂 Base path for input data."
    ),
    path_out: Path = typer.Option(
        "data/v-profiles",
        exists=True,
        writable=True,
        help="📂 Base path for output data."
    ),
    apriori_cfg: Path = typer.Option(
        "config",
        help="📂 Base path for a priori config file."
    ),
    
    update_data: bool = typer.Option(True, help="📈 Update data."),
    update_calendar: bool = typer.Option(True, help="🗓️ Update calendar."),
    update_map: bool = typer.Option(True, help="🗺️ Update map."),
    update_climatology: bool = typer.Option(True, help="↪️ Update climatology."),
    progress_bar: bool = typer.Option(True, help="⌛ Show progress bar."),
):
    """
    run aprofiles standard workflow for given dates and specific instruments types
    
    see some examples [here](https://augustinmortier.github.io/a-profiles/cli/)
    """

    #typer.echo(f"dates: {dates}, today: {today}, yesterday: {yesterday}, from: {_from}, to: {_to}, instruments_types: {instruments_types}, multiprocessing: {multiprocessing}")

    #prepare dates array | convert from tuples to list
    dates = list(_dates)
    # today / yesterday
    if today:
        dates.append(datetime.today())
    if yesterday:
        dates.append(datetime.today() - timedelta(days=1))
    # from / to
    if _from is not None:
        for date in date_range(_from, _to, freq="D"):
            dates.append(date)
    # abort if no dates
    if (len(dates) == 0):
        print("No provided date. Please check aprocess --help for more information.")
        raise typer.Abort()

    # progress bar
    if progress_bar:
        disable_progress_bar = False
    else:
        disable_progress_bar = True

    # read config file
    CFG = utils.config.read()
    
    # insert apriori config path
    for instrument_type in CFG["parameters"]:
        apriori = CFG["parameters"][instrument_type]["inversion"]["apriori"]
        if "cfg" in apriori:
            CFG["parameters"][instrument_type]["inversion"]["apriori"]["cfg"] = str(Path(apriori_cfg, apriori["cfg"]))

    for date in dates:
        yyyy = str(date.year)
        mm = str(date.month).zfill(2)
        dd = str(date.day).zfill(2)

        # list all files in in directory
        datepath = Path(path_in, yyyy, mm, dd)
        onlyfiles = [str(e) for e in datepath.iterdir() if e.is_file()]

        # data processing
        if update_data:
            if multiprocessing:
                with Progress() as progress:
                    task = progress.add_task(f"{date.strftime('%Y-%m-%d')} :rocket:", total=len(onlyfiles), visible=not disable_progress_bar)
                    with concurrent.futures.ProcessPoolExecutor(max_workers=workers) as executor:
                        futures = [executor.submit(
                            utils.workflow.workflow, 
                            path=file, 
                            instruments_types=instruments_types, 
                            base_dir=path_out, CFG=CFG, verbose=False
                        )
                        for file in onlyfiles]
                        for future in concurrent.futures.as_completed(futures):
                            progress.update(task, advance=1)
            else:
                for file in track(onlyfiles, description=f"{date.strftime('%Y-%m-%d')}   ", disable=disable_progress_bar):
                    utils.workflow.workflow(
                        file, instruments_types, path_out, CFG, verbose=False
                    )
        
        # list all files in out directory
        datepath = Path(path_out, yyyy, mm, dd)

        if update_calendar:
            # create calendar
            calname = f"{yyyy}-{mm}-cal.json"
            path = Path(path_out, yyyy, mm, calname)
            if not path.is_file():
                utils.calendar.make_calendar(path_out, yyyy, mm, calname)

            # list all files in out directory
            onlyfiles = [str(e) for e in datepath.iterdir() if e.is_file()]
            # add to calendar
            for file in track(onlyfiles, description="calendar     ", disable=disable_progress_bar):
                utils.calendar.add_to_calendar(file, path_out, yyyy, mm, dd, calname)
            
        
        if update_map:
            # create map
            mapname = f"{yyyy}-{mm}-map.json"
            path = Path(path_out, yyyy, mm, mapname)
            if not path.is_file():
                utils.map.make_map(path_out, yyyy, mm, mapname)

            # list all files in out directory
            onlyfiles = [str(e) for e in datepath.iterdir() if e.is_file()]
            # add to map
            for file in track(onlyfiles, description="map          ", disable=disable_progress_bar):
                utils.map.add_to_map(file, path_out, yyyy, mm, dd, mapname)

    if update_climatology:
        # list all files in out directory
        onlyfiles = [str(e) for e in datepath.iterdir() if e.is_file()]
        # get station id from file name
        stations_id = ["-".join(onlyfile.split("/")[-1].split("AP_")[1].split("-", 5)[:5]) for onlyfile in onlyfiles]
        # exclude moving stations
        stations_id = [station for station in stations_id if station not in CFG["exclude_stations_id_from_climatology"]]
        
        if multiprocessing:
            with Progress() as progress:
                task = progress.add_task(total=len(stations_id), description=f"clim.      :rocket:", visible=not disable_progress_bar)
                with concurrent.futures.ProcessPoolExecutor(max_workers=workers) as executor:
                    futures = [executor.submit(
                        utils.climatology.compute_climatology,
                        path_out, 
                        station_id, 
                        season_variables=["extinction"],
                        all_variables=["aod", "lidar_ratio"],
                        aerosols_only=True
                    )
                    for station_id in stations_id]
                    for future in concurrent.futures.as_completed(futures):
                        progress.update(task, advance=1)
        else:
            for station_id in track(stations_id, description='clim.        ', disable=disable_progress_bar):
                utils.climatology.compute_climatology(path_out, station_id, season_variables=["extinction"], all_variables=["aod", "lidar_ratio"], aerosols_only=True)


@app.command()
def l2b(
        path_in: Path = typer.Option(
            "data/v-profiles", exists=True, readable=True, help="📂 Base path for input data."
        ),
        path_out: Path = typer.Option(
            "data/l2b", exists=True, writable=True, help="📂 Base path for output data."
        ),
        time_steps: int = typer.Option(
            6, help="🔂 Number of most recent time steps to be processed."
        ),
        progress_bar: bool = typer.Option(True, help="⌛ Show progress bar.")
    ):
    """
    make E-PROFILE L2b files out of AP files
    """

    # if path_in is "data/v-profiles", use today's date to find the directory
    if path_in == Path("data/v-profiles"):
        # get todays date
        today = datetime.today()
        path_in = Path(path_in, today.strftime('%Y'), today.strftime('%m'), today.strftime('%d'))
    
    utils.l2b.make_files(path_in, path_out, time_steps, progress_bar)

if __name__ == "__main__":
    app()
